#include <stdbool.h>
#include <string.h>

#define RT_DONT_DEFINE_RENDERER_GLOBAL
#include "gfx.h"
#include "renderer_api.h"

#include "runtime/config.h"
#include "runtime/dynamic_libs.h"

/* Attributes are used to bind buffers (or textures) to symbolic values.
 * For example, an attribute might be bound to "CELL_GRID", which would be
 * replaced with the (at the time of the invoke) grid buffer of the current
 * world cell.
 */

rt_renderer_api g_renderer;
extern rt_cvar rt_MaxFramegraphs;

#ifndef RT_STATIC_LIB
static rt_dynlib _renderer_lib;
#endif
static bool _renderer_loaded = false;

RT_DLLEXPORT
RT_CVAR_S(rt_Renderer, "Select the render backend. Available options: [vk, null], Default: vk", "vk");

#ifdef RT_STATIC_LIB
extern void RT_RENDERER_API_FN(RegisterCVars)(void);
extern rt_result RT_RENDERER_API_FN(Init)(const rt_renderer_init_info *);
extern void RT_RENDERER_API_FN(Shutdown)(void);
extern void RT_RENDERER_API_FN(BeginFrame)(unsigned int);
extern void RT_RENDERER_API_FN(EndFrame)(unsigned int);
extern rt_pipeline_handle RT_RENDERER_API_FN(CompilePipeline)(const rt_pipeline_info *);
extern void RT_RENDERER_API_FN(DestroyPipeline)(rt_pipeline_handle);
extern rt_render_target_handle
    RT_RENDERER_API_FN(CreateRenderTarget)(const rt_render_target_info *);
extern rt_render_target_handle RT_RENDERER_API_FN(GetSwapchainRenderTarget)(void);
extern void RT_RENDERER_API_FN(DestroyRenderTarget)(rt_render_target_handle);
extern rt_result RT_RENDERER_API_FN(AllocCommandBuffers)(uint32_t,
                                                         const rt_alloc_command_buffer_info *,
                                                         rt_command_buffer_handle *);
extern rt_result RT_RENDERER_API_FN(SubmitCommandBuffers)(rt_gpu_queue,
                                                          const rt_submit_command_buffers_info *);
extern rt_result RT_RENDERER_API_FN(CreateSemaphores)(uint32_t,
                                                      const rt_gpu_semaphore_info *,
                                                      rt_gpu_semaphore_handle *);
extern void RT_RENDERER_API_FN(DestroySemaphores)(uint32_t count, rt_gpu_semaphore_handle *);
extern uint64_t RT_RENDERER_API_FN(GetSemaphoreValue)(rt_gpu_semaphore_handle);
extern rt_gpu_semaphore_handle RT_RENDERER_API_FN(GetSwapchainAvailableSemaphore)(void);
extern rt_gpu_semaphore_handle RT_RENDERER_API_FN(GetRenderFinishedSemaphore)(void);
extern void RT_RENDERER_API_FN(CmdBeginPass)(rt_command_buffer_handle,
                                             const rt_cmd_begin_pass_info *);
extern void RT_RENDERER_API_FN(CmdEndPass)(rt_command_buffer_handle);
extern void RT_RENDERER_API_FN(CmdTransitionRenderTarget)(rt_command_buffer_handle,
                                                          rt_render_target_handle,
                                                          rt_render_target_state);
#endif

extern rt_result InitFramegraphManager(void);
extern void ShutdownFramegraphManager(void);
extern rt_result InitRenderLists(void);
extern void ShutdownRenderLists(void);
extern void ResetRenderLists(void);

static bool LoadRenderer(void) {

#if !defined(RT_STATIC_LIB)
#define RETRIEVE_SYMBOL(name, type)                                                                \
    g_renderer.name = (type *)rtGetSymbol(_renderer_lib, "rtRen" #name);                           \
    if (!g_renderer.name) {                                                                        \
        rtReportError("GFX",                                                                       \
                      "Unable to retrieve renderer function %s from backend %s",                   \
                      #name,                                                                       \
                      rt_Renderer.s);                                                              \
    }

    if (strcmp(rt_Renderer.s, "vk") == 0) {
        _renderer_lib = rtOpenLib(RT_DLLNAME("rtvk"));
        if (!_renderer_lib) {
            rtReportError("GFX", "Unable to load renderer backend: %s", RT_DLLNAME("rtvk"));
            return false;
        }
        RETRIEVE_SYMBOL(RegisterCVars, rt_register_renderer_cvars_fn);
        RETRIEVE_SYMBOL(Init, rt_init_renderer_fn);
        RETRIEVE_SYMBOL(Shutdown, rt_shutdown_renderer_fn);
        RETRIEVE_SYMBOL(BeginFrame, rt_begin_frame_fn);
        RETRIEVE_SYMBOL(EndFrame, rt_end_frame_fn);
        RETRIEVE_SYMBOL(CompilePipeline, rt_compile_pipeline_fn);
        RETRIEVE_SYMBOL(DestroyPipeline, rt_destroy_pipeline_fn);
        RETRIEVE_SYMBOL(CreateRenderTarget, rt_create_render_target_fn);
        RETRIEVE_SYMBOL(GetSwapchainRenderTarget, rt_get_swapchain_render_target_fn);
        RETRIEVE_SYMBOL(DestroyRenderTarget, rt_destroy_render_target_fn);
        RETRIEVE_SYMBOL(AllocCommandBuffers, rt_alloc_command_buffers_fn);
        RETRIEVE_SYMBOL(SubmitCommandBuffers, rt_submit_command_buffers_fn);
        RETRIEVE_SYMBOL(CreateSemaphores, rt_create_gpu_semaphores_fn);
        RETRIEVE_SYMBOL(DestroySemaphores, rt_destroy_gpu_semaphores_fn);
        RETRIEVE_SYMBOL(GetSemaphoreValue, rt_get_gpu_semaphore_value_fn);
        RETRIEVE_SYMBOL(GetSwapchainAvailableSemaphore, rt_get_swapchain_available_semaphore_fn);
        RETRIEVE_SYMBOL(GetRenderFinishedSemaphore, rt_get_render_finished_semaphore_fn);
        RETRIEVE_SYMBOL(CmdBeginPass, rt_cmd_begin_pass_fn);
        RETRIEVE_SYMBOL(CmdEndPass, rt_cmd_end_pass_fn);
        RETRIEVE_SYMBOL(CmdTransitionRenderTarget, rt_cmd_transition_render_target_fn);
    } else {
        rtReportError("GFX",
                      "Unsupported renderer backend: (%s) %s",
                      rt_Renderer.name,
                      rt_Renderer.s);
        return false;
    }
#undef RETRIEVE_SYMBOL
#else
    g_renderer.RegisterCVars                  = &rtRenRegisterCVars;
    g_renderer.Init                           = &rtRenInit;
    g_renderer.Shutdown                       = &rtRenShutdown;
    g_renderer.BeginFrame                     = &rtRenBeginFrame;
    g_renderer.EndFrame                       = &rtRenEndFrame;
    g_renderer.CompilePipeline                = &rtRenCompilePipeline;
    g_renderer.DestroyPipeline                = &rtRenDestroyPipeline;
    g_renderer.CreateRenderTarget             = &rtRenCreateRenderTarget;
    g_renderer.GetSwapchainRenderTarget       = &rtRenGetSwapchainRenderTarget;
    g_renderer.DestroyRenderTarget            = &rtRenDestroyRenderTarget;
    g_renderer.AllocCommandBuffers            = &rtRenAllocCommandBuffers;
    g_renderer.SubmitCommandBuffers           = &rtRenSubmitCommandBuffers;
    g_renderer.CreateSemaphores               = &rtRenCreateSemaphores;
    g_renderer.DestroySemaphores              = &rtRenDestroySemaphores;
    g_renderer.GetSemaphoreValue              = &rtRenGetSemaphoreValue;
    g_renderer.GetSwapchainAvailableSemaphore = &rtRenGetSwapchainAvailableSemaphore;
    g_renderer.GetRenderFinishedSemaphore     = &rtRenGetRenderFinishedSemaphore;
    g_renderer.CmdBeginPass                   = &rtRenCmdBeginPass;
    g_renderer.CmdEndPass                     = &rtRenCmdEndPass;
    g_renderer.CmdTransitionRenderTarget      = &rtRenCmdTransitionRenderTarget;
#endif
    return true;
}

RT_DLLEXPORT void rtRegisterRendererCVars(void) {
    if (!_renderer_loaded) {
        if (!LoadRenderer())
            return;
        _renderer_loaded = true;
    }
    g_renderer.RegisterCVars();
}

RT_DLLEXPORT rt_result rtInitGFX(rt_renderer_init_info *renderer_info) {
    rtRegisterCVAR(&rt_Renderer);
    rtRegisterCVAR(&rt_MaxFramegraphs);

    if (!_renderer_loaded) {
        if (!LoadRenderer())
            return RT_UNKNOWN_ERROR;
        g_renderer.RegisterCVars();
    }

    rt_result result;

    if ((result = g_renderer.Init(renderer_info)) != RT_SUCCESS)
        return result;

    if ((result = InitFramegraphManager()) != RT_SUCCESS)
        return result;

    if ((result = InitRenderLists()) != RT_SUCCESS)
        return result;

    return result;
}

RT_DLLEXPORT void rtShutdownGFX(void) {
    ShutdownRenderLists();
    ShutdownFramegraphManager();
    g_renderer.Shutdown();
}

RT_DLLEXPORT void rtBeginGFXFrame(unsigned int frame_id) {
    g_renderer.BeginFrame(frame_id);
}

RT_DLLEXPORT void rtEndGFXFrame(unsigned int frame_id) {
    g_renderer.EndFrame(frame_id);
    ResetRenderLists();
}
