#include "render_list.h"

#include "runtime/threading.h"
#include "runtime/mem_arena.h"
#include "runtime/config.h"

#include <string.h>

RT_CVAR_I(rt_RenderListPoolSize,
          "Size of the pool allocated for render lists in bytes. Default: 8 MiB",
          RT_MB(8));

typedef struct {
    size_t size;
    const char *name;
} rt_render_object_type_data;

typedef struct rt_list_pool_s {
    size_t capacity;
    struct rt_list_pool_s *next;
} rt_list_pool;

#define DEFAULT_LIST_CAPACITY RT_KB(1)

static rt_render_object_type_data _types[RT_MAX_RENDER_OBJECT_TYPE + 1];
static unsigned int _type_count = 0;
static rt_rwlock _type_lock;

static rt_arena _list_arena;
static rt_list_pool *_first_free_list;
static rt_mutex *_list_lock;

rt_result InitRenderLists(void) {
    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok)
        return RT_UNKNOWN_ERROR;
    _type_lock = lock_res.lock;
 
    rt_create_arena_result arena_res = rtCreateArena(NULL, (size_t)rt_RenderListPoolSize.i);
    if (!arena_res.ok) {
        rtDestroyRWLock(&_type_lock);
        return RT_OUT_OF_MEMORY;
    }
    _list_arena = arena_res.arena;

    _list_lock = rtCreateMutex();
    if (!_list_lock) {
        rtReleaseArena(&_list_arena);
        rtDestroyRWLock(&_type_lock);
    }

    return RT_SUCCESS;
}

void ShutdownRenderLists(void) {
    rtDestroyRWLock(&_type_lock);
    rtDestroyMutex(_list_lock);
    rtReleaseArena(&_list_arena);
}

RT_DLLEXPORT  rt_render_object_type rtRegisterRenderObjectType(size_t object_size,
                                                              const char *debug_name) {
    if (_type_count == RT_MAX_RENDER_OBJECT_TYPE) {
        rtReportError("GFX", "Too many render object types (max is %u)", RT_MAX_RENDER_OBJECT_TYPE);
        return RT_INVALID_RENDER_OBJECT_TYPE;
    }
    rtLockWrite(&_type_lock);
    rt_render_object_type type = (rt_render_object_type)++_type_count;
    _types[_type_count].size = object_size;
    _types[_type_count].name = debug_name;
    if (debug_name)
        rtLog("GFX",
              "Registered render object type %s; object size: %zu. Type: %u",
              debug_name,
              object_size,
              _type_count);
    else
        rtLog("GFX",
              "Registered unnamed render object type; object size: %zu. Type: %u",
              object_size,
              _type_count);
    rtUnlockWrite(&_type_lock);
    return type;
}


RT_DLLEXPORT size_t rtGetRenderObjectSize(rt_render_object_type type) {
    size_t size = 0;
    rtLockRead(&_type_lock);
    if (type > RT_INVALID_RENDER_OBJECT_TYPE && type <= _type_count)
        size = _types[type].size;
    rtUnlockRead(&_type_lock);
    return size;
}

RT_DLLEXPORT const char *rtGetRenderObjectTypeDebugName(rt_render_object_type type) {
    const char *name = NULL;
    rtLockRead(&_type_lock);
    if (type > RT_INVALID_RENDER_OBJECT_TYPE && type <= _type_count)
        name = _types[type].name;
    rtUnlockRead(&_type_lock);
    return name;
}

static rt_create_render_list_result CreateNewList(rt_render_object_type type, size_t capacity) {
    rt_create_render_list_result res = {.ok = false};
    rtLockMutex(_list_lock);

    if (!_first_free_list || _first_free_list->capacity < capacity) { /* Allocate a new list */
        rt_list_pool *pool =
            rtArenaPush(&_list_arena, sizeof(rt_list_pool) + capacity);
        if (!pool) {
            rtReportError("GFX",
                          "Out of render list pool space! Configured space: %d kiB",
                          rt_RenderListPoolSize.i / 1024);
            goto out;
        }
        pool->capacity   = capacity;
        pool->next       = _first_free_list;
        _first_free_list = pool;
    }
    rt_render_list list;
    list.data        = (char *)_first_free_list + sizeof(rt_list_pool);
    list.type        = type;
    list.length      = 0;
    res.ok           = true;
    res.list         = list;
    _first_free_list = _first_free_list->next;
out:
    rtUnlockMutex(_list_lock);
    return res;
}

RT_DLLEXPORT rt_create_render_list_result rtCreateRenderList(rt_render_object_type type) {
    return CreateNewList(type, DEFAULT_LIST_CAPACITY);
}

void ResetRenderLists(void) {
    rtLockMutex(_list_lock);
    _first_free_list = NULL;
    rtArenaClear(&_list_arena);
    rtUnlockMutex(_list_lock);
}

RT_DLLEXPORT bool rtPushRenderListEntry(rt_render_list *list, const void *object) {
    size_t object_size = rtGetRenderObjectSize(list->type);
    rt_list_pool *pool = (rt_list_pool *)((char *)list->data - sizeof(rt_list_pool));
    size_t list_capacity = pool->capacity / object_size;

    if (list->length == list_capacity) {
        /* "Grow" the list */
        rt_create_render_list_result list_res = CreateNewList(list->type, pool->capacity * 2);
        if (!list_res.ok)
            return false;

        memcpy(list_res.list.data, list->data, list->length * object_size);
        
        rtLockMutex(_list_lock);
        pool->next = _first_free_list;
        _first_free_list = pool;
        rtUnlockMutex(_list_lock);

        list_res.list.length = list->length;
        *list = list_res.list;
    }

    char *dst = (char *)list->data + list->length * object_size;
    memcpy(dst, object, object_size);
    ++list->length;
    return true;
}