#ifndef RT_RENDER_LIST_H
#define RT_RENDER_LIST_H

/* a render list collects render objects. */

#include <stdint.h>
#include <stdbool.h>

#include "runtime/runtime.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Identifies a type of render objects. */
typedef uint32_t rt_render_object_type;
typedef uint32_t rt_render_object_type_mask;

enum {
    RT_INVALID_RENDER_OBJECT_TYPE = 0,
    RT_MAX_RENDER_OBJECT_TYPE     = 32,
};

#define RT_RENDER_OBJECT_TYPE_BIT(type) (1u << ((type)-1))

/* Registers a new render object type.
 * debug_name is optional and may be NULL.
 */
RT_DLLEXPORT rt_render_object_type rtRegisterRenderObjectType(size_t object_size,
                                                              const char *debug_name);

RT_DLLEXPORT size_t rtGetRenderObjectSize(rt_render_object_type type);

RT_DLLEXPORT const char *rtGetRenderObjectTypeDebugName(rt_render_object_type type);

typedef struct {
    rt_render_object_type type;
    size_t length;
    void *data;
} rt_render_list;

/* Returns a pointer to the i-th render list element.
 * Works for every valid type, because the size is determined at runtime. */
RT_INLINE void *rtGetRenderListElement(const rt_render_list *list, size_t index) {
    size_t size = rtGetRenderObjectSize(list->type);
    return (char *)list->data + size * index;
}

/* Returns the i-th render list element, cast to type T.
 * Saves a rtGetRenderObjectSize call, if the type is known beforehand. */
#define RT_GET_RENDER_LIST_ELEMENT(list, T, index) *(((T *)(list).data) + (index))

typedef struct {
    bool ok;
    rt_render_list list;
} rt_create_render_list_result;

/* Create a render list for a particular object type.
 *
 * Render Lists have a lifetime of one frame. */
RT_DLLEXPORT rt_create_render_list_result rtCreateRenderList(rt_render_object_type type);

/* Append a render object to a list. The object must be of the correct type. */
RT_DLLEXPORT bool rtPushRenderListEntry(rt_render_list *list, const void *object);

#ifdef __cplusplus
}
#endif

#endif
